/*
 * mouth.c - Kedrigern (Milena TTS server for Raspberry Pi)
 * Copyright (C) Bohdan R. Rau 2019 <ethanak@polip.com>
 *
 * Kedrigern is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * Kedrigern is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with Kedrigern.  If not, write to:
 * 	The Free Software Foundation, Inc.,
 * 	51 Franklin Street, Fifth Floor
 * 	Boston, MA  02110-1301, USA.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <ctype.h>
#include <wiringPiSPI.h>
#include <unistd.h>

#include "kedrigern_internal.h"


/* max7219 */
void matrix_Write(int matrix, uint8_t adr, uint8_t data)
{
    uint8_t buf[2];
    buf[0] = adr;
    buf[1] = data;
    write(matrix, buf, 2);
}

void matrix_Brightness(int matrix, int brightness)
{
    matrix_Write(matrix, 0x0a, brightness);
}

void matrix_Mode(int matrix, int mode)
{
    matrix_Write(matrix, 0x09, mode);
}

void matrix_Scanlimit(int matrix, int limit)
{
    matrix_Write(matrix, 0x0b, limit);
}

void matrix_PowerUp(int matrix, int mode)
{
    matrix_Write(matrix, 0x0c, mode);
}

void matrix_Shape(int matrix, uint8_t *shape)
{
    int i;
    for (i=0; i<8; i++) matrix_Write(matrix, i+1,shape[i]);
}

void matrix_TestMode(int matrix, int mode)
{
    matrix_Write(matrix, 0x0f, mode);
}
/* end 7219 */

void rotate(uint8_t *in, uint8_t *out, int mode)
{
    int x, y;
    int get_point(int x, int y)
    {
        return (in[x] & (1<<y)) ? 1 : 0;
    }
    void set_point(int x,int y) {
        out[x] |= 1<<y;
    }
    
    switch(mode) {
        case 1:
        for (x=0; x< 8; x++) for (y=0; y < 8; y++) {
            if (get_point(y,7-x)) set_point(x,y);
        }
        break;
        
        case 2:
        for (x=0; x< 8; x++) for (y=0; y < 8; y++) {
            if (get_point(7-x,7-y)) set_point(x,y);
        }
        break;
        
        case 3:
        for (x=0; x< 8; x++) for (y=0; y < 8; y++) {
            if (get_point(7-y,x)) set_point(x,y);
        }
        break;
        
        default:
        memcpy(out, in, 8);
    }
}

uint8_t paszcza_in[BANK_COUNT * PHONE_COUNT][8]={

// bank 0
    {0,0,0,0,0,0,0},  // BLANK
    {0xc3,0,0,0x3c,0x42,0x81,0x42,0x2c},  // A
    {0xc3,0,0,0x7e,0x81,0x81,0x81,0x7e},  // E
    {0xc3,0,0,0x7e,0x81,0x81,0x7e,0},  // I
    {0xc3,0,0,0x3c,0x42,0x42,0x42,0x3c},  // O
    //{0xc3,0,0x18,0x24,0x24,0x18,0},  // U
    {0xc3,0,0,0x18,0x24,0x18,0},  // U
    {0xc3,0,0,0x7e,0x81,0x42,0x3c,0},  // Y
    {0xc3,0,0,0,0xff,0x7e,0,0},  // P
    {0xc3,0,0,0,0xff,0x7e,0,0},  // M
    {0xc3,0,0,0x7e,0x81,0x7e,0},  // F
    {0xc3,0,0,0x3c,0x42,0x24,0x18,0},  // W
    {0xc3,0,0,0x3c,0x42,0x3c,0},  // ETC
    {0xc3,0,0,0,0x3c,0,0,0},  // SIL
    {0xc3,0,0,0,0x66,0x3c,0,0},  // REST
// bank 1 (bez oczek)
    {0,0,0,0,0,0,0},  // BLANK
    {0,0,0x3c,0x42,0x81,0x42,0x2c},  // A
    {0,0,0x7e,0x81,0x81,0x81,0x7e},  // E
    {0,0,0x7e,0x81,0x81,0x7e,0},  // I
    {0,0,0x3c,0x42,0x42,0x42,0x3c},  // O
    {0,0,0x18,0x24,0x18,0},  // U
    {0,0,0x7e,0x81,0x42,0x3c,0},  // Y
    {0,0,0,0xff,0x7e,0,0},  // P
    {0,0,0,0xff,0x7e,0,0},  // M
    {0,0,0x7e,0x81,0x7e,0},  // F
    {0,0,0x3c,0x42,0x24,0x18,0},  // W
    {0,0,0x3c,0x42,0x3c,0},  // ETC
    {0,0,0,0x7e,0,0,0},  // SIL
    {0,0,0x66,0x3c,0,0}  // REST



    };

uint8_t shapaszcza[PHONE_COUNT][8];
uint32_t shapeset;

uint8_t paszcza[PHONE_COUNT][8];

int servomouth[PHONE_COUNT] = {
    0,
    80, // A
    100, // E
    50, // I
    60, // O
    40, // U
    25, // Y
    0, // P
    0, // M
    10, // F
    30, // W
    20, // ETC
    10, // SIL
    0, // REST
};
    
static int matrix;
static int servo, servostart, servoend;


static struct mouthem *mouth, *mitem;
static int mouth_len, mouth_size, mouth_time, mouth_pos;

static const char * milephon = "aVbeWpfiJmoYuOwvyEUI";
static const uint8_t phonon[] = {
    PHONE_A,PHONE_Y,PHONE_P,PHONE_E,0,PHONE_P,PHONE_F,PHONE_I,0,
    PHONE_M,PHONE_O,PHONE_E,PHONE_U,PHONE_O,PHONE_W,PHONE_F,
    PHONE_Y,PHONE_E,PHONE_U,PHONE_Y};

static struct mouthem *add_mouthem(uint8_t shape, int time, int stress)
{
    if (!mouth) {
        mouth = malloc(sizeof(*mouth) * (mouth_size = 1024));
        mouth_len=0;
    }
    else if (mouth_len >= mouth_size) {
        mouth_size *= 2;
        mouth = realloc(mouth, sizeof(*mouth) * (mouth_size = 1024));
    }
    mouth[mouth_len].shape = shape;
    mouth[mouth_len].time = time;
    mouth[mouth_len].stress = stress;
    
    return mouth + (mouth_len++);
}


static void add_line(char *line)
{
    int ti, pho=0, stres=0;
    if (line[0] == ';' && line[1] == 'M') {
        line += 2;
        while (*line && isspace(*line)) line++;
        if (!*line) return;
        char *c=strchr(milephon,*line);
        if (!c) pho = PHONE_ETC;
        else pho = phonon[c-milephon];
        stres=strtol(line+1, NULL, 10)+1;
        ti=0;
        
    }
    else if (line[0]=='_') {
        ti=strtol(line+1, NULL, 10);
        pho = PHONE_SIL;
    }
    else {
        pho=0;
        while (*line && !isspace(*line)) line++;
        ti = strtol(line,NULL, 10);
    }
    if (!pho && !mitem) {
        mitem = add_mouthem(PHONE_SIL,ti, 0);
        return;
    }
    if (!pho) {
        mitem->time += ti;
        return;
    }
    if (mitem && mitem->shape == PHONE_P && (pho == PHONE_P || pho == PHONE_M)) {
        int t = mitem->time;
        mitem -> time = (t * 2) / 3;
        mitem = add_mouthem(PHONE_ETC, t - mitem->time, 0);
    }
    else if (pho == PHONE_SIL && mitem && mitem->shape == PHONE_SIL) {
        mitem->time += ti;
        return;
    }
    mitem = add_mouthem(pho, ti, stres);
}

static int nodev;

void startMouth(char *pho, double factor)
{
    if (nodev) return;
    mitem = NULL;
    mouth_len = 0;
    mouth_pos = 0;
    while(pho) {
        while (*pho && isspace(*pho)) pho++;
        add_line(pho);
        pho = strchr(pho,'\n');
    }
    int i,ti = 0, tj=0;
    
    for (i=0; i<mouth_len; i++) {
        int t = mouth[i].time;
        mouth[i].time = ti * factor;
        if (i) tj = mouth[i-1].time;
        ti += t;
    }
    if (servo < 0) matrix_PowerUp(matrix,1);
}


void initServo(int s, int s_start, int s_end)
{
    servo = s;
    servostart = s_start;
    servoend = s_end;
}

void moveServo(int pos)
{
    int p = servostart + (pos * (servoend - servostart)) / 100;
    FILE *f=fopen("/dev/servoblaster","w");
    if (f) {
        fprintf(f,"%d=%d\n", servo, p);
        fclose(f);
    }
}

void shapeServo(int shape)
{
    moveServo(servomouth[shape]);
}

static uint8_t decodeShapeLine(char *c)
{
    int i;
    uint8_t rc=0;
    for (i=0;i<8;i++) {
        if (!c[i]) break;
        if (!isspace(c[i]) && c[i] !='.') rc |= 128 >> i;
    }
    return rc;
}

static void readShape(const char *shape)
{
    FILE *f=fopen(shape,"r");
    char *s, *s1, buf[256];
    char *ts[]={"ETC", "SIL","REST","A","E","I","O","U","Y","P","F","W",NULL};
    int i, nshape, j;
    uint8_t cs[]={PHONE_ETC, PHONE_SIL, PHONE_REST,
        PHONE_A, PHONE_E, PHONE_I, PHONE_O, PHONE_U, PHONE_Y,
        PHONE_P, PHONE_F, PHONE_W, 0};
        
    char *despace(char *c)
    {
        while (*c && isspace(*c)) c++;
        return c;
    }
    if (!f) {
        perror(shape);
        return;
    }
    while (fgets(buf,256,f)) {
        s=despace(buf);
        if (!*buf == '#') continue;
        if (strncasecmp(s,"shape",5) || !isspace(s[5])) continue;
        s=despace(s+5);
        for (i=0;ts[i];i++) {
            if (!strncasecmp(s,ts[i],strlen(ts[i]))) {
                s=despace(s+strlen(ts[i]));
                if (!*s || *s == '#') break;
            }
        }
        if (!ts[i]) continue;
        nshape = cs[i];
        for (j=0;j<8;j++) {
            if (!fgets(buf,256,f)) break;
            shapaszcza[nshape][j] = decodeShapeLine(buf);
        }
        if (nshape == PHONE_P) memcpy(shapaszcza[PHONE_M],shapaszcza[PHONE_P],8);
    }
    fclose(f);
}


void initMouth(int dev, int rot, int bri, int nd,
    int serv, int s_start, int s_end, int ablank,
    int bank, const char *shape)
{
    if (nodev = nd) return;
    if (serv >= 0) {
        initServo(serv, s_start, s_end);
        return;
    }
    else {
        servo = -1;
    }
    matrix = wiringPiSPISetup(dev,500000);
    if (matrix < 0) {
        perror("Bład inicjalizacji SPI");
        exit(1);
    }
    matrix_Brightness(matrix, bri);
    matrix_Mode(matrix, 0);
    matrix_Scanlimit(matrix,7);
    matrix_PowerUp(matrix, ablank ? 0 :1);
    matrix_TestMode(matrix, 0);
    bank = constrain(bank, 0, BANK_COUNT-1);
    memcpy(shapaszcza, paszcza_in[PHONE_COUNT],8*PHONE_COUNT);
    if (shape) {
        readShape(shape);
    }
    for (int i = 0; i < PHONE_COUNT; i++) {
        rotate(shapaszcza[i], paszcza[i], rot);
    }
    matrix_Shape(matrix, paszcza[PHONE_REST]);
}

void shapeMouth(int shape)
{
    if (nodev) return;
    if (servo >= 0) shapeServo(shape);
    else matrix_Shape(matrix, paszcza[shape]);
}

void setMouth(int ti)
{
    if (nodev) return;
    int z = 0;
    if (ti < 0) {
        if (servo >= 0) shapeServo(PHONE_REST);
        else matrix_Shape(matrix, paszcza[PHONE_REST]);
        return;
    }
    for (;;) {
        if (mouth_pos >= mouth_len) break;
        if (mouth[mouth_pos].time >= ti+10) break;
        z = mouth[mouth_pos++].shape;
    }
    if (z) {
        if (servo >= 0) shapeServo(z);
        else matrix_Shape(matrix, paszcza[z]);
    }

}

void finalizeMouth()
{
    if (nodev) return;
    if (servo >= 0) shapeServo(PHONE_REST);
    else matrix_Shape(matrix, paszcza[PHONE_REST]);
}

void brightnessMouth(int b)
{
    if (nodev) return;
    if (servo < 0) matrix_Brightness(matrix, b);
}

void powerMouth(int b)
{
    if (servo < 0) matrix_PowerUp(matrix, b ? 1 : 0);
}

void phonemMouth(int shape)
{
    if (shape < 0 || shape > PHONE_REST) return;
    if (nodev) return;
    if (servo >= 0) shapeServo(shape);
    else {
        powerMouth(1);
        matrix_Shape(matrix, paszcza[shape]);
    }
}
